import disamby.preprocessors as prep
from disamby import Disamby
import pytest


pipeline = [prep.normalize_whitespace,
            prep.remove_punctuation,
            prep.compact_abbreviations,
            lambda x: prep.ngram(x, 4)]


@pytest.mark.parametrize('size', [20, 1000, 2000])
def test_fitting(size, company_df, benchmark):
    df = company_df(size)
    benchmark(Disamby, df, pipeline)


@pytest.mark.parametrize('size', [20, 1000, 2000])
def test_sparse_find(size, company_df, benchmark):
    df = company_df(size)
    dis = Disamby(df, pipeline)
    idx = list(dis.records['name'].keys())[0]
    results = benchmark(dis.find, idx, .7)
    score_of_searched = max(x.score for x in results)
    assert score_of_searched == pytest.approx(1)




import os
import pytest
import pandas as pd
import numpy as np
from faker import Faker


TEST_DIR = os.path.dirname(__file__)
DATA_DIR = os.path.join(TEST_DIR, 'data')


@pytest.fixture
def fake_names():
    """function creating fake names of size n"""
    def f_names(n):
        fake = Faker()
        fake.seed(n)
        names = [fake.address() for _ in range(n)]
        return names
    return f_names


@pytest.fixture
def company_df(fake_names):
    """test company details"""
    def df_of_size(size):
        np.random.seed(size)
        assignee_path = os.path.join(DATA_DIR, 'potential_assginees_names.csv')
        sample = pd.read_csv(assignee_path, index_col='inv_id').sample(size)
        sample['address'] = fake_names(size)
        del sample['patents']
        return sample
    return df_of_size



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `disamby` package."""
import pytest
from pytest import mark

from disamby.preprocessors import compact_abbreviations
from disamby.preprocessors import split_words
from disamby.preprocessors import normalize_whitespace
from disamby.preprocessors import remove_punctuation
from disamby.preprocessors import ngram
from disamby.preprocessors import nword


@mark.parametrize('raw,expected', [
    ('this is a   long  string', 'THIS IS A LONG STRING'),
    ('this has a \tb', 'THIS HAS A B',)
])
def test_reduce_duplicate_whitespace(raw, expected):
    assert normalize_whitespace(raw) == expected


@mark.parametrize('raw,expected', [
    ('this is i.b.m', 'THIS IS IBM'),
    ('an other A.B.M this', 'AN OTHER ABM THIS')
])
def test_compact_abbreviations(raw, expected):
    assert compact_abbreviations(raw) == expected


@mark.parametrize('raw,n,expected', [
    ('this', 2, ('th', 'hi', 'is')),
    ('this', 3, ('thi', 'his')),
    ('this', 4, ('this',)),
    ('this', 5, ('this',)),
    ('this', 1, ValueError),
    ('this', -3, ValueError)
])
def test_ngram(raw, n, expected):
    if expected is ValueError:
        with pytest.raises(expected):
            ngram(raw, n)
    else:
        assert ngram(raw, n) == expected


def test_split_words():
    words = split_words("this is a sentence")
    assert len(words) == 4

    words = split_words("this")
    assert len(words) == 1


def test_remove_punctuation():
    assert remove_punctuation('.has a .few!') == 'has a few'


@mark.parametrize('raw,expected', [
    ('this Ias I.B.M',
     ('THIS', 'HIS ', 'IS I', 'S IA', ' IAS', 'IAS ', 'AS I', 'S IB', ' IBM')
     )
])
def test_combined_preprocessors(raw, expected):
    reduced = normalize_whitespace(raw)
    abbreviated = compact_abbreviations(reduced)
    trigram = ngram(abbreviated, 4)
    assert trigram == expected


@pytest.mark.parametrize('raw,k,expected',[
    ('this that the other', 2, ('thisthat', 'thatthe', 'theother')),
    ('this that the other', 4, ('thisthattheother',)),
    ('this that the other', 20, ('thisthattheother',))
])
def test_word_sequencer(raw, k, expected):
    assert nword(raw, k) == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `disamby` package."""
import pytest
import pandas as pd
from disamby import Disamby
import disamby.preprocessors as prep


pipeline = [
    prep.normalize_whitespace,
    prep.remove_punctuation,
    prep.compact_abbreviations,
    lambda x: prep.ngram(x, 4)
]


@pytest.fixture
def disamby_fitted_instance(fake_names):
    names = fake_names(90)
    data_series = pd.Series(names)
    dis = Disamby()
    dis.fit(data_series, pipeline, 'streets')
    return dis



def test_frequency_counter(disamby_fitted_instance):
    dis = disamby_fitted_instance
    assert 'streets' in dis.fields
    counter = dis.field_freq['streets']
    assert counter.most_common(1)[0][1] >= 1


def test_identification_potential(disamby_fitted_instance):
    dis = disamby_fitted_instance

    weights = dis.id_potential(('st', 'street', 'suite'), 'streets')
    assert sum(weights.values()) == pytest.approx(1)


@pytest.mark.parametrize('smoother,offset,expected', [
    (None, 0, 0),
    ('offset', 1000, 0),
    ('log', 10000, 0),
])
def test_scoring(smoother, offset, expected, disamby_fitted_instance):
    dis = disamby_fitted_instance
    score = dis.score('David Heights', 'Rebecca Shoal Suite', 'streets',
                      smoother=smoother, offset=offset)
    assert score == pytest.approx(expected, abs=.01)


def test_instant_instantiation(company_df):
    df = company_df(500)
    dis = Disamby(df, pipeline)
    dis.field_freq['address'].most_common()


def test_find(company_df):
    df = company_df(100)
    dis = Disamby(df, pipeline)
    term = list(dis.records['address'].keys())[0]
    results = dis.find(term, threshold=0.001, weights={'name': .2, 'address': .8})

    assert len(results) == 42
    score_of_searched = max(x.score for x in results)
    assert score_of_searched == pytest.approx(1)


def test_disambiguated_tests(company_df):
    df = company_df(200)
    dis = Disamby(df, preprocessors=pipeline)
    components = dis.disambiguated_sets(verbose=True, threshold=.7,
                                        weights={'name': .99, 'address': .01}
                                        )
    assert max(len(c) for c in components) == 2



from click.testing import CliRunner
from disamby import cli


def test_command_line_interface():
    """Test the CLI."""
    # TODO: use temporary directory to do actual testing
    runner = CliRunner()
    result = runner.invoke(cli.main, [])
    assert result.exit_code == 2  # missing parameters
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert 'Show this message and exit.' in help_result.output



# -*- coding: utf-8 -*-

"""Console script for disamby.

With the cli it is possible to carry out the disambiguation on the command line.
The script returns a json file with a name taken from the first field as the key and
a list of indices belonging to the same disambiguated cluster.

.. code-block:: python

    {'International Business Machines Inc': [1, 2, 5, 34, 90],
    'Samsung': [4, 123],
    '...': [...],
     ...}

To carry out the disambiguation you will need a csv file with proper headings
and optionally an index column. If not index column is specified then the position in the
csv is used in the json file to identify its members.

Only a basic processing pipeline is possible here, see the `--help` of the script
to know which are available.

Example usage:

.. code-block:: bash

    $ disamby --col-headers name,address \\
                --index inv_id \\
                --threshold .7 \\
                --prep APNX \\
                input.csv output.json

"""

import click
from disamby import Disamby
import disamby.preprocessors as pre
import pandas as pd
import json

# TODO: Make tests for cli


@click.command(help="""
This command allows to disambiguate a csv file using `disamby`
example usage:\n
$ disamby -c name -i inv_id -t .7 -p APNX name.csv dis_name.json
$ disamby -c name -i inv_id -t .7 -p APNX name.csv dis_name.json

The prep codes are:\n
'A': pre.compact_abbreviations\n
'P': pre.remove_punctuation\n
'W': pre.normalize_whitespace\n
'3': lambda x: pre.ngram(x, 3)\n
'4': lambda x: pre.ngram(x, 4)\n
'5': lambda x: pre.ngram(x, 5)\n
'S': pre.split_words\n
'X': lambda x: pre.ngram(x[:33], 4) + pre.split_words(x)
""")
@click.argument('data', type=click.Path(dir_okay=False, exists=True))
@click.argument('output', type=click.Path(dir_okay=False))
@click.option('-c', '--col-headers', type=click.STRING,
              help='name of the columns containing the strings to match if more then '
                   'one then separate with comma (no spaces)')
@click.option('-t', '--threshold', type=click.FLOAT, default=0.7,
              help='Minimum score to be counted')
@click.option('-i', '--index', type=click.STRING, help='name of index column')
@click.option('-p', '--prep', type=click.STRING, default='APWX',
              help='pre-processing instructions, if left blank will default to a standard'
                   ' prep')
def main(data, output, index, col_headers, prep, threshold):
    """Console script for disamby."""
    names_df = pd.read_csv(data, index_col=index)
    prep_dict = {
        'A': pre.compact_abbreviations,
        'P': pre.remove_punctuation,
        'W': pre.normalize_whitespace,
        '3': lambda x: pre.ngram(x, 3),
        '4': lambda x: pre.ngram(x, 4),
        '5': lambda x: pre.ngram(x, 5),
        'S': pre.split_words,
        'X': lambda x: pre.ngram(x[:33], 4) + pre.split_words(x)
    }
    columns = col_headers.split(',')
    pipeline = [prep_dict[action] for action in list(prep)]
    dis = Disamby(data=names_df[columns], preprocessors=pipeline)
    components = dis.disambiguated_sets(threshold, smoother='offset', offset=100)

    comp_to_id = dict()
    for comp in components:
        members = list(comp)
        representative = members[0]
        name = names_df.loc[representative, columns[0]]
        comp_to_id[name] = members

    with open(output, 'w') as f:
        json.dump(comp_to_id, f)


if __name__ == '__main__':
    main()





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=6.0',
    'networkx',
    'tqdm'
]

setup_requirements = [
    'pytest-runner'
]

test_requirements = [
    'pytest',
    'pandas',
    'faker',
]

setup(
    name='disamby',
    version='0.2.2',
    description="Python package to carry out entity disambiguation based on "
                "string matching",
    long_description=readme + '\n\n' + history,
    author="Luca Verginer",
    author_email='luca@verginer.eu',
    url='https://github.com/verginer/disamby',
    packages=find_packages(include=['disamby']),
    entry_points={
        'console_scripts': [
            'disamby=disamby.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords=['disamby', 'disambiguation'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



# -*- coding: utf-8 -*-

"""
This module contains the various string preprocessors
"""
import re

__all__ = ['compact_abbreviations', 'normalize_whitespace',
           'ngram', 'trigram', 'split_words', 'remove_punctuation',
           'nword']


_re_abbreviation = re.compile(r'\.(?![a-zA-Z]*})')
_re_duplicate_white = re.compile('\s+')
_re_whitespace = re.compile('\s')
_re_punctuation = re.compile('[^\w\s]')


def normalize_whitespace(string: str) -> str:
    """
    removes duplicates whitespaces as well as replace tabs and newlines with a space

    Parameters
    ----------
    string

    Returns
    -------
    str

    Examples
    --------
    >>> normalize_whitespace('this is a  \t long  string')
    'THIS IS A LONG STRING'
    """
    no_whitespace = _re_duplicate_white.sub(' ', string.upper())
    return no_whitespace.strip()


def compact_abbreviations(string: str) -> str:
    """
    Removes dots between single letters and concatenates them

    Parameters
    ----------
    string

    Returns
    -------
    str

    Examples
    --------
    >>> compact_abbreviations('an other A.B.M this')
    'AN OTHER ABM THIS'
    """
    split = _re_abbreviation.split(string.upper())
    return ''.join(split)


def remove_punctuation(word: str) -> str:
    """
    removes all punctuation symbols from the string

    Parameters
    ----------
    word: str

    Returns
    -------
    str

    Examples
    --------
    >>> remove_punctuation('.has -a .few!')
    'has a few'
    """
    return _re_punctuation.sub('', word)


def ngram(string: str, n: int) -> tuple:
    """
    constructs all possible ngrams from the given string. If the string is shorter then
    the n then the string is returned

    Parameters
    ----------
    string
    n : int
        value must be larger at least 2

    Returns
    -------
    tuple of strings

    Examples
    --------
    >>> ngram('this', 2)
    ('th', 'hi', 'is')
    """
    N = len(string)
    if n > N:
        return string,
    if n < 2:
        raise ValueError('n for a ngram must be 2 or larger')
    return tuple(string[i:i+n] for i in range(N-n+1))


def trigram(string: str) -> tuple:
    return ngram(string, 3)


def split_words(string: str) -> tuple:
    """
    splits words on whitespace. This function is more reliable then `.split(' ')`
    since it works with any whitespace character (i.e. those recognized by regex)

    Parameters
    ----------
    string

    Returns
    -------
    tuple of strings

    Examples
    --------
    >>> len(split_words('a new day'))
    3
    """
    return tuple(_re_whitespace.split(string))


def nword(word: str, k: int) -> tuple:
    """
    concatenates k consecutive words into a tuple

    Parameters
    ----------
    word
    k

    Returns
    -------
    tuple of strings

    Examples
    --------
    >>> nword('this that the other', 2)
    ('thisthat', 'thatthe', 'theother')
    """
    parts = split_words(word)
    n = len(parts)
    k = min(k, n)
    sequences = tuple(''.join(parts[i:i + k]) for i in range(n - k + 1))
    return sequences



# -*- coding: utf-8 -*-

"""Main module."""
from collections import Counter
from typing import Union
from pandas import DataFrame, Series
from math import log
from tqdm import tqdm
from networkx import DiGraph
from collections import namedtuple

ScoredElement = namedtuple('ScoredElement', ['index', 'score'])
PandasObj = Union[DataFrame, Series]


class Disamby(object):
    """
    Class for disambiguation fitting, scoring and ranking of potential matches

    A `Disamby` instance stores the pre-processing pipeline applied to the
    strings for a given field as well as the the computed frequencies from the
    entire corpus of strings to match against.
    `Disamby` can be instantiated either with not arguments, with a list of
    strings, pandas.Series or pandas.DataFrame. This triggers the immediate
    call to the `fit` method, whose doc explains the parameters.

    Examples
    --------
    >>> import pandas as pd
    >>> import disamby.preprocessors as pre
    >>> df = pd.DataFrame(
    ... {'a': ['Luca Georger', 'Luca Geroger', 'Adrian Sulzer'],
    ... 'b': ['Mira, 34, Augsburg', 'Miri, 34, Augsburg', 'Milano, 34']
    ... }, index=['L1', 'L2', 'O1']
    ... )
    >>> pipeline = [
    ...     pre.normalize_whitespace,
    ...     pre.remove_punctuation,
    ...     pre.trigram
    ... ]
    >>> dis = Disamby(df, pipeline)
    >>> dis.disambiguated_sets(threshold=0.5, verbose=False)
    [{'L2', 'L1'}, {'O1'}]
    """

    def __init__(self, data: PandasObj=None, preprocessors: list=None, field: str=None):
        self.field_freq = dict()
        self.preprocessors = dict()
        self.records = dict()
        self._processed_token_cache = dict()
        self._most_common = dict()
        self._token_to_instance = dict()
        if data is not None:
            if preprocessors is None:
                raise ValueError("Preprocessor not provided")
            self.fit(data, preprocessors, field)

    def fit(self, data: PandasObj, preprocessors: list, field: str=None):
        """
        Computes the frequencies of the terms by field.

        Parameters
        ----------
        data : pandas.DataFrame, pandas.Series or list of strings
            list of strings or pandas.DataFrame
            if dataframe is given then the field defaults to the column name
        preprocessors : list
            list of functions to apply in that order
            note the first function must accept a string, the other functions
            must be such that a pipeline is possible the result is a tuple of
            strings.
        field : str
            string identifying which field this data belongs to

        Examples
        --------

        >>> import pandas as pd
        >>> from disamby.preprocessors import split_words
        >>> df = pd.DataFrame(
        ... {'a': ['Luca Georger', 'Luke Geroge', 'Adrian Sulzer'],
        ... 'b': ['Mira, 34, Augsburg', 'Miri, 32', 'Milano, 34']
        ... })
        >>> dis = Disamby()
        >>> prep = [split_words]
        >>> dis.fit(df, prep)
        """
        try:
            columns = data.columns
            for col in columns:
                self._fit_field(data[col], preprocessors=preprocessors)
        except AttributeError:
            self._fit_field(data, preprocessors=preprocessors, field=field)

    def find(self, idx, threshold=0.0, weights: dict=None, **kwargs) -> list:
        """
        returns the list of scored instances which have a score above the
        threshold. Note that strings which do not share any token are omitted
        since their score is 0 by default.

        Parameters
        ----------
        idx
            index of the record to find
        threshold
        weights : dict

        Returns
        -------

        """
        fields = self.fields

        scored_candidates = dict()
        if weights is None:
            weights = {f: 1 / len(fields) for f in fields}

        for field in fields:
            own_term = self.records[field][idx]
            own_tokens = self._processed_token_cache[field][own_term]
            potential_candidates = set()

            for token in own_tokens:
                potential_candidates |= self._token_to_instance[field][token]

            for candidate in potential_candidates:
                score = self.score(own_term, candidate[1], field, **kwargs)
                candidate_idx = candidate[0]
                if candidate_idx not in scored_candidates:
                    scored_candidates[candidate_idx] = {field: score * weights[field]}
                else:
                    scored_candidates[candidate_idx][field] = score * weights[field]

        # compute weighted score
        final_candidates = []
        for idx, scores in scored_candidates.items():
            total_score = sum(scores.values())
            if total_score >= threshold:
                final_candidates.append(ScoredElement(idx, total_score))

        return final_candidates

    def _fit_field(self, data: PandasObj, preprocessors: list=None, field: str=None):
        if field not in self.preprocessors:
            ValueError('preprocessors have already been defined, '
                       'cannot redefine. This would render the lookup '
                       'inconsistent')

        if field is None:
            try:
                field = data.name
            except AttributeError:  # was not a pandas.Series
                raise KeyError("The provided data are not a pandas Series, "
                               "if the data is a list you need to provide the"
                               "`field` argument.")

        self.preprocessors[field] = preprocessors
        self._processed_token_cache[field] = dict()
        self._token_to_instance[field] = dict()
        self.records[field] = dict()
        counter = Counter()

        for i, name in data.items():
            norm_tokens = self.pre_process(name, preprocessors)
            self._processed_token_cache[field][name] = norm_tokens
            counter.update(norm_tokens)
            self.records[field][i] = name
            for token in norm_tokens:
                if token in self._token_to_instance[field]:
                    self._token_to_instance[field][token] |= {(i, name)}
                else:
                    self._token_to_instance[field][token] = {(i, name)}
        self._most_common[field] = counter.most_common(1)[0][1]
        self.field_freq[field] = counter

    def score(self, term: str, other_term: str, field: str,
              smoother=None, offset=0) -> float:
        """
        Computes the score between the two strings using the frequency data

        Parameters
        ----------
        term : str
            term to search for
        other_term : str
            the other term to compare too
        field : str
            the name of the column to which this term belongs
        smoother : str (optional)
            one of {None, 'offset', 'log'}
        offset : int
            offset to add to count only needed for smoothers 'log' and 'offset'

        Returns
        -------
        float

        Notes
        -----
        The score is not commutative (i.e. score(A,B) != score(B,A))
        """

        try:
            own_parts = self._processed_token_cache[field][term]
            other_parts = self._processed_token_cache[field][other_term]
        except KeyError:
            own_parts = self.pre_process(term, self.preprocessors[field])
            other_parts = self.pre_process(other_term, self.preprocessors[field])

        # get list of potential scores
        weights = self.id_potential(own_parts, field, smoother, offset)
        score = sum(weights.get(tok, 0) for tok in other_parts)
        return score

    def id_potential(self, term: Union[tuple, str], field: str,
                     smoother: str = None, offset=0) -> dict:
        """
        Computes the weights of the words based on the observed frequency
        and normalized.

        Parameters
        ----------
        term : str, tuple
            term to look for or a tuple of proper tokens
        field : str
            field the word falls under
        smoother : str (optional)
            one of {None, 'offset', 'log'}
        offset : int
            offset to add to count only needed for smoothers 'log' and 'offset'

        Returns
        -------
        float
        """
        if isinstance(term, str):
            words = self.pre_process(term, self.preprocessors[field])
        else:
            words = term

        smoothers = {
            None: self._smooth_none,
            'offset': self._smooth_offset,
            'log': self._smooth_log
        }
        if smoother not in smoothers:
            raise KeyError(
                'Chosen smother `{}` is not one of {}'.format(
                    smoother, smoothers.keys())
            )
        counter = self.field_freq[field]

        s_fun = smoothers[smoother]
        max_occ = self._most_common[field]
        id_potentials = {
            word: s_fun(counter[word], offset, max_occ)
            for word in words
        }

        total_weight = sum(id_potentials.values())
        return {w: idp / total_weight for w, idp in id_potentials.items()}

    def alias_graph(self, threshold=0.7, verbose=True, weights=None, **kwargs) -> DiGraph:
        """
        This function creates the directed network connecting an instance to an other
        through a directed edge if the the target instance has a similarity score above
        the threshold.

        Parameters
        ----------
        weights
        threshold : float
            between 0 and 1
        verbose : whether to show the progressbar
        kwargs :
            arguments to pass to the score function (i.e. offset, smoother)

        Returns
        -------
        DiGraph

        """
        if not verbose:
            def t(x):
                return x
        else:
            t = tqdm

        edges = []
        fields = self.records.keys()
        a_field = list(fields)[0]
        for idx in t(self.records[a_field]):
            targets = self.find(idx, threshold=threshold, weights=weights, **kwargs)
            new_edges = [(idx, x.index, {'score': x.score}) for x in targets]
            edges.extend(new_edges)

        alias_graph = DiGraph()
        alias_graph.add_edges_from(edges)
        return alias_graph

    def disambiguated_sets(self, threshold=0.7, verbose=True, weights=None, **kwargs):
        from networkx import strongly_connected_components
        alias_graph = self.alias_graph(threshold, verbose, weights, **kwargs)
        strong = strongly_connected_components(alias_graph)
        return list(strong)

    @staticmethod
    def pre_process(base_name, functions: list):
        """apply every function consecutively to base_name"""
        norm_name = base_name
        for f in functions:
            norm_name = f(norm_name)
        return set(norm_name)

    @property
    def fields(self):
        return self.field_freq.keys()

    @staticmethod
    def _smooth_none(occurrences, *args):
        return 1 / max(occurrences, 1)

    @staticmethod
    def _smooth_offset(occurrences, offset, *args):
        return 1 / max(occurrences + offset, 1)

    @staticmethod
    def _smooth_log(occ, offset, max_occ):
        max_offset = max(max_occ + offset + 1, 1)
        word_offset = max(occ + offset, 1)
        return log(max_offset / word_offset)
